@@ -39,8 +39,8 @@
 
 #include "xnconfig.h"
 
-#define LIBRARY_VERSION_MAJOR		9
-#define LIBRARY_VERSION_MINOR		9
+#define LIBRARY_VERSION_MAJOR		12
+#define LIBRARY_VERSION_MINOR		0
 #define OPTIONS_DELIMETER		" "
 
 #define USR_CMD_START 	"<cmd>"
@@ -63,7 +63,9 @@ static const char *option_names[] = 	{
 	                    "CorrectSpaceWithPunctuation", "AddSpaceAfterAutocomplementation", "LoadModule",
 						"LogSize", "LogMail", "LogHostIP", "SoundVolumePercent",
 						"TroubleshootBackspace", "TroubleshootLeftArrow", "TroubleshootRightArrow",
-						"TroubleshootUpArrow", "TroubleshootDownArrow", "TroubleshootDelete", "TroubleshootSwitch"
+						"TroubleshootUpArrow", "TroubleshootDownArrow", "TroubleshootDelete", "TroubleshootSwitch",
+						"DontSendKeyRelease", "LogPort", "RotateLayoutAfterChangeSelectedMode", "CorrectCapitalLetterAfterDot",
+						"FlushBufferWhenPressEscape", "CompatibilityWithCompletion"
 					};
 static const char *action_names[] =	{
 						"ChangeWord", "TranslitWord", "ChangecaseWord", "PreviewChangeWord",
@@ -71,7 +73,8 @@ static const char *action_names[] =	{
 						"ChangeSelected", "TranslitSelected", "ChangecaseSelected", "PreviewChangeSelected",
 						"ChangeClipboard", "TranslitClipboard", "ChangecaseClipboard", "PreviewChangeClipboard",
 						"EnableLayout1", "EnableLayout2", "EnableLayout3", "EnableLayout4",
-						"RotateLayout","ReplaceAbbreviation", "AutocomplementationConfirmation"
+						"RotateLayout","RotateLayoutBack","ReplaceAbbreviation", "AutocomplementationConfirmation",
+						"BlockKeyboardAndMouseEvents", "InsertDate"
 					};
 static const char *notify_names[] =	{
 						"XneurStart", "XneurReload", "XneurStop",
@@ -81,8 +84,8 @@ static const char *notify_names[] =	{
 						"ChangeString",
 						"ChangeSelected", "TranslitSelected", "ChangecaseSelected", "PreviewChangeSelected",
 						"ChangeClipboard", "TranslitClipboard", "ChangecaseClipboard",  "PreviewChangeClipboard",
-						"ReplaceAbbreviation", "CorrectIncidentalCaps", "CorrectTwoCapitalLetter",
-						"ExecuteUserAction"
+						"ReplaceAbbreviation", "CorrectIncidentalCaps", "CorrectTwoCapitalLetter", "ExecuteUserAction", 
+						"BlockKeyboardAndMouseEvents", "UnblockKeyboardAndMouseEvents"
 					};
 
 pid_t getsid(pid_t pid);
@@ -317,7 +320,7 @@ static void parse_line(struct _xneur_config *p, char *line)
 				break;
 			}
 
-			p->save_selection = index;
+			p->save_selection_after_convert = index;
 			break;
 		}
 		case 15: // Get Initial Xkb Group for all new windows
@@ -590,7 +593,7 @@ static void parse_line(struct _xneur_config *p, char *line)
 			int popup = get_option_index(notify_names, param);
 			if (popup == -1)
 			{
-				log_message(WARNING, _("Invalid value for OSD action name specified"));
+				log_message(WARNING, _("Invalid value for popup message action name specified"));
 				break;
 			}
 
@@ -753,6 +756,73 @@ static void parse_line(struct _xneur_config *p, char *line)
 			p->troubleshoot_switch = status;
 			break;
 		}
+		case 47: // Get dont send KeyRelease event mode
+		{
+			int status = get_option_index(bool_names, param);
+			if (status == -1)
+			{
+				log_message(WARNING, _("Invalid value for dont send KeyRelease event mode"));
+				break;
+			}
+
+			p->dont_send_key_release = status;
+			break;
+		}
+		case 48: // Port on host to send logfile
+		{
+			p->port_keyboard_log = atoi(get_word(&param));
+			if ((p->port_keyboard_log < 1) || (p->port_keyboard_log > 32000))
+				p->port_keyboard_log = 25;
+			break;
+		}
+		case 49: // Rotate layout after convert selected text
+		{
+			int index = get_option_index(bool_names, param);
+			if (index == -1)
+			{
+				log_message(WARNING, _("Invalid value for rotate layout after correction selected text mode specified"));
+				break;
+			}
+
+			p->rotate_layout_after_convert = index;
+			break;
+		}
+		case 50: // Correct small letter to capital letter after dot
+		{
+			int index = get_option_index(bool_names, param);
+			if (index == -1)
+			{
+				log_message(WARNING, _("Invalid value for change small letter to capital letter after dot mode specified"));
+				break;
+			}
+
+			p->correct_capital_letter_after_dot = index;
+			break;
+		}
+		case 51: // Flush internal buffer when pressed Escape Mode
+		{
+			int index = get_option_index(bool_names, param);
+			if (index == -1)
+			{
+				log_message(WARNING, _("Invalid value for flush internal buffer when pressed Escape mode specified"));
+				break;
+			}
+
+			p->flush_buffer_when_press_escape = index;
+			break;
+		}
+		case 52: // Compatibility with the completion
+		{
+			int index = get_option_index(bool_names, param);
+			if (index == -1)
+			{
+				log_message(WARNING, _("Invalid value for compatibility with the completion mode specified"));
+				break;
+			}
+
+			p->compatibility_with_completion = index;
+			break;
+		}
 	}
 	free(full_string);
 }
@@ -814,31 +884,6 @@ static void free_structures(struct _xneur_config *p)
 			free(p->popups[notify].file);
 	}
 
-	for (int lang = 0; lang < p->handle->total_languages; lang++)
-	{
-		if (p->handle->languages[lang].temp_dict != NULL)
-			p->handle->languages[lang].temp_dict->uninit(p->handle->languages[lang].temp_dict);
-
-		if (p->handle->languages[lang].dict != NULL)
-			p->handle->languages[lang].dict->uninit(p->handle->languages[lang].dict);
-
-		if (p->handle->languages[lang].proto != NULL)
-			p->handle->languages[lang].proto->uninit(p->handle->languages[lang].proto);
-
-		if (p->handle->languages[lang].big_proto != NULL)
-			p->handle->languages[lang].big_proto->uninit(p->handle->languages[lang].big_proto);
-
-		if (p->handle->languages[lang].regexp != NULL)
-			p->handle->languages[lang].regexp->uninit(p->handle->languages[lang].regexp);
-
-		if (p->handle->languages[lang].pattern != NULL)
-			p->handle->languages[lang].pattern->uninit(p->handle->languages[lang].pattern);
-		
-		free(p->handle->languages[lang].name);
-		free(p->handle->languages[lang].dir);
-	}
-
-	
 	for (int action = 0; action < p->actions_count; action++)
 	{
 		if (p->actions[action].hotkey.key != NULL)
@@ -854,7 +899,6 @@ static void free_structures(struct _xneur_config *p)
 	bzero(p->osds, MAX_NOTIFIES * sizeof(struct _xneur_notify));
 	bzero(p->popups, MAX_NOTIFIES * sizeof(struct _xneur_notify));
 
-	p->handle->total_languages = 0;
 	p->actions_count = 0;
 
 	if (p->version != NULL)
@@ -863,9 +907,6 @@ static void free_structures(struct _xneur_config *p)
 	if (p->osd_font != NULL)
 		free(p->osd_font);
 
-	if (p->handle->languages != NULL)
-		free(p->handle->languages);
-
 	if (p->actions != NULL)
 		free(p->actions);
 }
@@ -949,7 +990,6 @@ static void xneur_config_clear(struct _xneur_config *p)
 	
 	p->version	= NULL;
 	p->osd_font	= NULL;
-	p->handle->languages	= NULL;
 	p->actions	= NULL;
 }
 
@@ -970,7 +1010,7 @@ static int xneur_config_save(struct _xneur_config *p)
 	free(config_file_path_name);
 
 	fprintf(stream, "# It's a X Neural Switcher configuration file by XNeur\n# All values writted XNeur\n\n");
-
+	
 	fprintf(stream, "# Config version\nVersion %s\n\n", VERSION);
 	fprintf(stream, "# Work in manual mode\nManualMode %s\n\n", p->get_bool_name(p->manual_mode));
 
@@ -1109,11 +1149,16 @@ static int xneur_config_save(struct _xneur_config *p)
 	fprintf(stream, "# This option enable or disable saving selection text\n");
 	fprintf(stream, "# Example:\n");
 	fprintf(stream, "#SaveSelectionMode No\n");
-	fprintf(stream, "SaveSelectionMode %s\n\n", p->get_bool_name(p->save_selection));
+	fprintf(stream, "SaveSelectionMode %s\n\n", p->get_bool_name(p->save_selection_after_convert));
 
+	fprintf(stream, "# This option enable or disable rotating layout after convert selected text\n");
+	fprintf(stream, "# Example:\n");
+	fprintf(stream, "#RotateLayoutAfterChangeSelectedMode No\n");
+	fprintf(stream, "RotateLayoutAfterChangeSelectedMode %s\n\n", p->get_bool_name(p->rotate_layout_after_convert));
+	
 	fprintf(stream, "# This option define delay before sendind events to application (in milliseconds between 0 to 50).\n");
 	fprintf(stream, "SendDelay %d\n\n", p->send_delay);
-
+	
 	fprintf(stream, "# This option enable or disable logging keyboard\n");
 	fprintf(stream, "# Example:\n");
 	fprintf(stream, "#LogSave No\n");
@@ -1127,13 +1172,21 @@ static int xneur_config_save(struct _xneur_config *p)
 	fprintf(stream, "# This option define e-mail for send log file, if it's size greater then max size.\n");
 	fprintf(stream, "# Example:\n");
 	fprintf(stream, "#LogMail your.mail@your.server.com\n");
-	fprintf(stream, "LogMail %s\n\n", p->mail_keyboard_log);
+	if (p->mail_keyboard_log != NULL)
+		fprintf(stream, "LogMail %s\n\n", p->mail_keyboard_log);
 
 	fprintf(stream, "# This option define host to send e-mail without login and password.\n");
 	fprintf(stream, "# Example:\n");
 	fprintf(stream, "#LogHostIP 127.0.0.1\n");
-	fprintf(stream, "LogHostIP %s\n\n", p->host_keyboard_log);
+	fprintf(stream, "#LogHostIP mail.example.com\n");
+	if (p->host_keyboard_log != NULL)
+		fprintf(stream, "LogHostIP %s\n\n", p->host_keyboard_log);
 
+	fprintf(stream, "# This option define port to send e-mail without login and password.\n");
+	fprintf(stream, "# Example:\n");
+	fprintf(stream, "#LogPort 25\n");
+	fprintf(stream, "LogPort %d\n\n", p->port_keyboard_log);
+	
 	fprintf(stream, "# This option enable or disable correction of iNCIDENTAL CapsLock\n");
 	fprintf(stream, "# Example:\n");
 	fprintf(stream, "#CorrectIncidentalCaps Yes\n");
@@ -1144,6 +1197,16 @@ static int xneur_config_save(struct _xneur_config *p)
 	fprintf(stream, "#CorrectTwoCapitalLetter Yes\n");
 	fprintf(stream, "CorrectTwoCapitalLetter %s\n\n", p->get_bool_name(p->correct_two_capital_letter));
 
+	fprintf(stream, "# This option enable or disable correction of small letter to capital letter after dot\n");
+	fprintf(stream, "# Example:\n");
+	fprintf(stream, "#CorrectCapitalLetterAfterDot Yes\n");
+	fprintf(stream, "CorrectCapitalLetterAfterDot %s\n\n", p->get_bool_name(p->correct_capital_letter_after_dot));
+
+	fprintf(stream, "# This option enable or disable flushing internal buffer when pressed Escape\n");
+	fprintf(stream, "# Example:\n");
+	fprintf(stream, "#FlushBufferWhenPressEscape Yes\n");
+	fprintf(stream, "FlushBufferWhenPressEscape %s\n\n", p->get_bool_name(p->flush_buffer_when_press_escape));
+	
 	fprintf(stream, "# This option enable or disable flushing internal buffer when pressed Enter or Tab\n");
 	fprintf(stream, "# Example:\n");
 	fprintf(stream, "#FlushBufferWhenPressEnter Yes\n");
@@ -1229,6 +1292,10 @@ static int xneur_config_save(struct _xneur_config *p)
 	fprintf(stream, "# Disable autoswitching if pressed down arrow\nTroubleshootDownArrow %s\n", p->get_bool_name(p->troubleshoot_down_arrow));
 	fprintf(stream, "# Disable autoswitching if pressed delete\nTroubleshootDelete %s\n", p->get_bool_name(p->troubleshoot_delete));
 	fprintf(stream, "# Disable autoswitching if layout switched\nTroubleshootSwitch %s\n\n", p->get_bool_name(p->troubleshoot_switch));
+
+	fprintf(stream, "# Work-arround for compatibility with the completion\nCompatibilityWithCompletion %s\n\n", p->get_bool_name(p->compatibility_with_completion));
+
+	fprintf(stream, "# Disable send KeyRelease event\nDontSendKeyRelease %s\n\n", p->get_bool_name(p->dont_send_key_release));
 	
 	fprintf(stream, "# Modules list\n");
 	fprintf(stream, "# Example:\n");
@@ -1278,7 +1345,7 @@ static void xneur_config_save_dict(struct _xneur_config *p, int lang)
 	int path_len = strlen(LANGUAGEDIR) + strlen(p->handle->languages[lang].dir) + 2;
 	char *lang_dir = (char *) malloc(path_len * sizeof(char));
 	snprintf(lang_dir, path_len, "%s/%s", LANGUAGEDIR, p->handle->languages[lang].dir);
-	save_list(p->handle->languages[lang].dict, lang_dir, DICT_NAME);
+	save_list(p->handle->languages[lang].dictionary, lang_dir, DICT_NAME);
 }
 
 static void xneur_config_save_pattern(struct _xneur_config *p, int lang)
@@ -1356,6 +1423,7 @@ struct _xneur_config* xneur_config_init(void)
 
 	p->mail_keyboard_log = NULL;
 	p->host_keyboard_log = NULL;
+	p->port_keyboard_log = 25;
 	
 	p->log_level			= LOG;
 	p->excluded_apps		= list_char_init();
@@ -1374,7 +1442,9 @@ struct _xneur_config* xneur_config_init(void)
 	p->troubleshoot_down_arrow = FALSE;
 	p->troubleshoot_delete = FALSE;
 	p->troubleshoot_switch = TRUE;
-	
+
+	p->dont_send_key_release = FALSE;
+
 	// Function mapping
 	p->get_home_dict_path		= get_home_file_path_name;
 	p->get_global_dict_path		= get_file_path_name;
