@@ -31,6 +31,9 @@
 #include <signal.h>
 #include <pthread.h>
 #include <ctype.h>
+#include <time.h>
+#include <unistd.h> 
+
 
 #ifdef WITH_ASPELL
 #  include <aspell.h>
@@ -60,8 +63,6 @@
 
 #include "notify.h"
 
-#include "math.h"
-
 #include "program.h"
 
 #define KLB_NO_ACTION           0	// Modifier, function etc
@@ -177,12 +178,12 @@ static int get_auto_action(struct _program *p, KeySym key, int modifier_mask)
 			return KLB_SPACE;
 	}
 
-	if (modifier_mask & Mod1Mask || modifier_mask & Mod4Mask)
-		return KLB_NO_ACTION;
+	/*if (modifier_mask & Mod1Mask || modifier_mask & Mod4Mask)
+		return KLB_NO_ACTION;*/
 
 	if (modifier_mask & ControlMask)
 		return KLB_CLEAR;
-	
+
 	return KLB_ADD_SYM;
 }
 
@@ -289,6 +290,8 @@ static void program_update(struct _program *p)
 	p->update_modifiers_stack(p);
 	
 	p->focus->update_events(p->focus, listen_mode);
+	// Сброс признака "ручное переключение" после смены фокуса.
+	p->changed_manual = MANUAL_FLAG_UNSET;
 }
 
 static void program_process_input(struct _program *p)
@@ -299,30 +302,43 @@ static void program_process_input(struct _program *p)
 	{
 		int type = p->event->get_next_event(p->event);
 
+		int curr_layout = get_curr_keyboard_group();
+		if (p->last_layout != curr_layout) 
+		{
+			p->last_layout = curr_layout;
+			if (xconfig->troubleshoot_switch)
+			{
+				//log_message (ERROR, "KBD_SWITCH") ;
+				p->changed_manual = MANUAL_FLAG_SET;
+			}		
+		}
+		
 		switch (type)
 		{
 			case ClientMessage:
 			{
 				// Exit from main cycle by message to main window
-				XClientMessageEvent *cme = (XClientMessageEvent *) &(p->event->event);
-				if (cme->message_type == main_window->close_atom)
-				{
-					log_message(LOG, _("Exitting from main cycle"));
-					return;
-				}
+				//XClientMessageEvent *cme = (XClientMessageEvent *) &(p->event->event);
+				//log_message(LOG, _("Exitting from main cycle"));
+				//return;
 				break;
 			}
 			case KeyPress:
 			{
+				if (xconfig->block_events)
+				{
+					XAllowEvents(main_window->display, AsyncKeyboard, CurrentTime);
+				}
+				
 				log_message(TRACE, _("Received KeyPress '%s' (event type %d)"), XKeysymToString(p->event->get_cur_keysym(p->event)), type);
-
+				
 				// Save received event
 				p->event->default_event = p->event->event;
 
 				// Processing received event
 				p->on_key_action(p, type);
 				
-				// Restore event
+				// Resend special key back to window
 				if (p->event->default_event.xkey.keycode != 0)
 				{
 					set_event_mask(p->focus->owner_window, None);
@@ -334,6 +350,11 @@ static void program_process_input(struct _program *p)
 			}
 			case KeyRelease:
 			{
+				if (xconfig->block_events)
+				{
+					XAllowEvents(main_window->display, AsyncKeyboard, CurrentTime);
+				}
+				
 				log_message(TRACE, _("Received KeyRelease '%s' (event type %d)"), XKeysymToString(p->event->get_cur_keysym(p->event)), type);
 
 				// Save received event
@@ -358,7 +379,8 @@ static void program_process_input(struct _program *p)
 			{
 				if (type == FocusIn)
 				{
-					log_message(TRACE, _("Received FocusIn (event type %d)"), type);
+					if (p->focus->owner_window != p->event->event.xfocus.window)
+						log_message(TRACE, _("Received FocusIn on window %d (event type %d)"), p->event->event.xfocus.window, type);
 
 					p->last_layout = get_curr_keyboard_group();
 
@@ -373,7 +395,8 @@ static void program_process_input(struct _program *p)
 			}
 			case FocusOut:
 			{
-				log_message(TRACE, _("Received FocusOut (event type %d)"), type);
+				if (p->focus->owner_window != p->event->event.xfocus.window)
+					log_message(TRACE, _("Received FocusOut on window %d (event type %d)"), p->event->event.xfocus.window, type);
 
 				p->last_layout = get_curr_keyboard_group();
 				p->update(p);
@@ -391,33 +414,40 @@ static void program_process_input(struct _program *p)
 			}
 			case ButtonPress:
 			{
-				p->buffer->save_and_clear(p->buffer, p->focus->owner_window);
-				log_message(TRACE, _("Received ButtonPress on window %d (event type %d)"), p->event->event.xbutton.subwindow, type);
-
-				// Unfreeze and resend grabbed event
-				XAllowEvents(main_window->display, ReplayPointer, CurrentTime);
-				//XAllowEvents(main_window->display, SyncPointer, CurrentTime);
+				// Clear buffer only when clicked left button
+				if (p->event->event.xbutton.button == Button1)
+				{
+					p->buffer->save_and_clear(p->buffer, p->focus->owner_window);
+					log_message(TRACE, _("Received Button%dPress on window %d with subwindow %d (event type %d)"), p->event->event.xbutton.button, p->event->event.xbutton.window, p->event->event.xbutton.subwindow, type);
+				}
 				
-				/*p->focus->update_events(p->focus, LISTEN_DONTGRAB_INPUT);
-				XTestFakeButtonEvent (main_window->display, 1, TRUE, CurrentTime);
-				p->focus->update_events(p->focus, LISTEN_GRAB_INPUT);*/
+				if (xconfig->block_events)
+				{
+					XAllowEvents(main_window->display, AsyncPointer, CurrentTime);
+					break;
+				}
+				
+				XAllowEvents(main_window->display, ReplayPointer, CurrentTime);
 
 				break;
 			}
 			case ButtonRelease:
-			{
-				p->buffer->save_and_clear(p->buffer, p->focus->owner_window);
-				log_message(TRACE, _("Received ButtonRelease on window %d (event type %d)"), p->event->event.xbutton.subwindow, type);
+			{	
+				// Clear buffer only when clicked left button
+				if (p->event->event.xbutton.button == Button1)
+				{
+					p->buffer->save_and_clear(p->buffer, p->focus->owner_window);
+				}
 
-				// Unfreeze and resend grabbed event
+				log_message(TRACE, _("Received Button%dRelease on window %d with subwindow %d (event type %d)"), p->event->event.xbutton.button, p->event->event.xbutton.window, type);
+				
+				if (xconfig->block_events)
+				{
+					XAllowEvents(main_window->display, AsyncPointer, CurrentTime);
+					break;
+				}
+				
 				XAllowEvents(main_window->display, ReplayPointer, CurrentTime);
-				//XAllowEvents(main_window->display, SyncPointer, CurrentTime);
-				
-				/*p->focus->update_events(p->focus, LISTEN_DONTGRAB_INPUT);
-				XTestFakeButtonEvent (main_window->display, 1, FALSE, CurrentTime);
-				p->focus->update_events(p->focus, LISTEN_GRAB_INPUT);*/
-				
-				break;
 			}
 			case PropertyNotify:
 			{
@@ -443,10 +473,15 @@ static void program_process_input(struct _program *p)
 				p->buffer = buffer_init(xconfig->handle);
 				main_window->keymap = keymap_init(xconfig->handle);
 				
-				log_message (DEBUG, "Now layouts count %d", xconfig->handle->total_languages);
+				log_message (DEBUG, _("Now layouts count %d"), xconfig->handle->total_languages);
 				p->update(p);
 				break;
 			}
+			case MotionNotify:
+			{
+				log_message(TRACE, _("Received MotionNotify (event type %d)"), type);
+				break;
+			}
 			default:
 			{
 				log_message(DEBUG, _("Uncatched event with type %d (see X11/X.h for details)"), type);
@@ -461,6 +496,7 @@ static void program_change_lang(struct _program *p, int new_lang)
 	log_message(DEBUG, _("Changing language from %s to %s"), xconfig->handle->languages[get_curr_keyboard_group()].name, xconfig->handle->languages[new_lang].name);
 	p->buffer->set_lang_mask(p->buffer, new_lang);
 	XkbLockGroup(main_window->display, XkbUseCoreKbd, new_lang);
+	p->last_layout = new_lang;
 }
 
 static void program_change_incidental_caps(struct _program *p)
@@ -517,7 +553,10 @@ static void program_process_selection_notify(struct _program *p)
 		case ACTION_CHANGE_SELECTED:
 		{
 			p->buffer->rotate_layout(p->buffer);
-
+			
+			if (xconfig->rotate_layout_after_convert)
+				set_next_keyboard_group(xconfig->handle);
+			
 			show_notify(NOTIFY_CHANGE_SELECTED, NULL);
 			break;
 		}
@@ -587,12 +626,16 @@ static void program_process_selection_notify(struct _program *p)
 
 	if (p->action_mode == ACTION_CHANGE_SELECTED || p->action_mode == ACTION_CHANGECASE_SELECTED || p->action_mode == ACTION_TRANSLIT_SELECTED)
 	{
-		if (xconfig->save_selection)
+		if (xconfig->save_selection_after_convert)
 			p->event->send_selection(p->event, p->buffer->cur_pos);
 	}
 
 	p->buffer->save_and_clear(p->buffer, p->focus->owner_window);
 
+	// Unblock keyboard
+	set_event_mask(p->focus->owner_window, INPUT_HANDLE_MASK | FOCUS_CHANGE_MASK | EVENT_KEY_MASK);
+	grab_spec_keys(p->focus->owner_window, TRUE);
+	
 	p->update(p);
 	p->action_mode = ACTION_NONE;
 }
@@ -602,22 +645,30 @@ static void program_update_modifiers_stack(struct _program *p)
 	// Update mask
 	p->prev_key_mod = 0;
 	
-	if (p->modifiers_stack->exist(p->modifiers_stack, "Shift_L", BY_PLAIN) || p->modifiers_stack->exist(p->modifiers_stack, "Shift_R", BY_PLAIN))
-		p->prev_key_mod += (1 << 0);
+	if (p->modifiers_stack->exist(p->modifiers_stack, "Shift_L", BY_PLAIN) || 
+	    p->modifiers_stack->exist(p->modifiers_stack, "Shift_R", BY_PLAIN))
+			p->prev_key_mod += (1 << 0);
 	if (p->modifiers_stack->exist(p->modifiers_stack, "Caps_Lock", BY_PLAIN))
-		p->prev_key_mod += (1 << 1);
-	if (p->modifiers_stack->exist(p->modifiers_stack, "Control_L", BY_PLAIN) || p->modifiers_stack->exist(p->modifiers_stack, "Control_R", BY_PLAIN))
-		p->prev_key_mod += (1 << 2);
-	if (p->modifiers_stack->exist(p->modifiers_stack, "Alt_L", BY_PLAIN) || p->modifiers_stack->exist(p->modifiers_stack, "Alt_R", BY_PLAIN))
-		p->prev_key_mod += (1 << 3);
-	if (p->modifiers_stack->exist(p->modifiers_stack, "Meta_L", BY_PLAIN) || p->modifiers_stack->exist(p->modifiers_stack, "Meta_R", BY_PLAIN))
-		p->prev_key_mod += (1 << 4);
+			p->prev_key_mod += (1 << 1);
+	if (p->modifiers_stack->exist(p->modifiers_stack, "Control_L", BY_PLAIN) || 
+	    p->modifiers_stack->exist(p->modifiers_stack, "Control_R", BY_PLAIN)
+	 	/* || p->modifiers_stack->exist(p->modifiers_stack, "ISO_Prev_Group", BY_PLAIN) || p->modifiers_stack->exist(p->modifiers_stack, "ISO_Next_Group", BY_PLAIN)*/)
+			p->prev_key_mod += (1 << 2);
+	if (p->modifiers_stack->exist(p->modifiers_stack, "Alt_L", BY_PLAIN) || 
+	    p->modifiers_stack->exist(p->modifiers_stack, "Alt_R", BY_PLAIN))
+			p->prev_key_mod += (1 << 3);
+	if (p->modifiers_stack->exist(p->modifiers_stack, "Meta_L", BY_PLAIN) || 
+	    p->modifiers_stack->exist(p->modifiers_stack, "Meta_R", BY_PLAIN))
+			p->prev_key_mod += (1 << 4);
 	if (p->modifiers_stack->exist(p->modifiers_stack, "Num_Lock", BY_PLAIN))
 		p->prev_key_mod += (1 << 5);
-	if (p->modifiers_stack->exist(p->modifiers_stack, "Super_L", BY_PLAIN) || p->modifiers_stack->exist(p->modifiers_stack, "Super_R", BY_PLAIN))
-		p->prev_key_mod += (1 << 6);
-	if (p->modifiers_stack->exist(p->modifiers_stack, "Hyper_L", BY_PLAIN) || p->modifiers_stack->exist(p->modifiers_stack, "Hyper_R", BY_PLAIN) || p->modifiers_stack->exist(p->modifiers_stack, "ISO_Level3_Shift", BY_PLAIN))
-		p->prev_key_mod += (1 << 7);
+	if (p->modifiers_stack->exist(p->modifiers_stack, "Super_L", BY_PLAIN) || 
+	    p->modifiers_stack->exist(p->modifiers_stack, "Super_R", BY_PLAIN) ||
+	    p->modifiers_stack->exist(p->modifiers_stack, "Hyper_L", BY_PLAIN) || 
+	    p->modifiers_stack->exist(p->modifiers_stack, "Hyper_R", BY_PLAIN))
+			p->prev_key_mod += (1 << 6);
+	if (p->modifiers_stack->exist(p->modifiers_stack, "ISO_Level3_Shift", BY_PLAIN))
+			p->prev_key_mod += (1 << 7);
 }
 
 static void program_on_key_action(struct _program *p, int type)
@@ -636,6 +687,13 @@ static void program_on_key_action(struct _program *p, int type)
 			p->modifiers_stack->add(p->modifiers_stack, keysym_str);
 		}
 		p->prev_key_mod |= p->event->get_cur_modifiers_by_keysym(p->event);
+
+		// If blocked events then processing stop 
+		if (xconfig->block_events)
+		{
+			p->event->default_event.xkey.keycode = 0;
+			return;
+		}
 		
 		int user_action = get_user_action(key, p->prev_key_mod);
 		enum _hotkey_action manual_action = get_manual_action(key, p->prev_key_mod);
@@ -648,6 +706,7 @@ static void program_on_key_action(struct _program *p, int type)
 		p->plugin->key_press(p->plugin, key, p->prev_key_mod);
 		
 		int auto_action = get_auto_action(p, key, p->prev_key_mod);
+	
 		if ((auto_action != KLB_NO_ACTION) && (auto_action != KLB_CLEAR))
 		{
 			int lang = get_curr_keyboard_group();
@@ -713,6 +772,7 @@ static void program_on_key_action(struct _program *p, int type)
 
 		p->modifiers_stack->rem(p->modifiers_stack, XKeysymToString(XK_Shift_R));
 		p->modifiers_stack->rem(p->modifiers_stack, XKeysymToString(XK_Shift_L));
+		p->modifiers_stack->rem(p->modifiers_stack, XKeysymToString(XK_ISO_Level3_Shift));
 		if (get_key_state(XK_Shift_R) != 0)
 			p->modifiers_stack->add(p->modifiers_stack, XKeysymToString(XK_Shift_R));
 		if (get_key_state(XK_Shift_L) != 0)
@@ -720,9 +780,19 @@ static void program_on_key_action(struct _program *p, int type)
 		if (get_key_state(XK_ISO_Level3_Shift) != 0)
 			p->modifiers_stack->add(p->modifiers_stack, XKeysymToString(XK_ISO_Level3_Shift));
 
-		
 		p->update_modifiers_stack(p);
 
+		// If blocked events then processing stop 
+		if (xconfig->block_events)
+		{
+			p->event->default_event.xkey.keycode = 0;
+			enum _hotkey_action manual_action = get_manual_action(key, modifier_mask);
+			if (manual_action == ACTION_BLOCK_EVENTS)
+				p->perform_manual_action(p, manual_action);
+
+			return;
+		}
+
 		p->plugin->key_release(p->plugin, key, p->prev_key_mod);
 		
 		int user_action = get_user_action(key, modifier_mask);
@@ -767,22 +837,25 @@ static void program_perform_user_action(struct _program *p, int action)
 static void program_perform_auto_action(struct _program *p, int action)
 {
 	struct _buffer *string = p->buffer;
-
 	switch (action)
 	{
 		case KLB_NO_ACTION:
 		{
-			if (!get_key_state(XK_Caps_Lock))
-				return;
-
-			if (!xconfig->disable_capslock)
-				return;
-
-			int xkb_opcode, xkb_event, xkb_error;
-			int xkb_lmaj = XkbMajorVersion;
-			int xkb_lmin = XkbMinorVersion;
-			if (XkbLibraryVersion(&xkb_lmaj, &xkb_lmin) && XkbQueryExtension(main_window->display, &xkb_opcode, &xkb_event, &xkb_error, &xkb_lmaj, &xkb_lmin))
-				XkbLockModifiers (main_window->display, XkbUseCoreKbd, LockMask, 0);
+			if (xconfig->flush_buffer_when_press_escape) 
+				if (p->event->get_cur_keysym(p->event) == XK_Escape)
+					p->buffer->save_and_clear(p->buffer, p->focus->owner_window);
+			
+			if (xconfig->disable_capslock)
+			{
+				if (!get_key_state(XK_Caps_Lock))
+					return;
+				
+				int xkb_opcode, xkb_event, xkb_error;
+				int xkb_lmaj = XkbMajorVersion;
+				int xkb_lmin = XkbMinorVersion;
+				if (XkbLibraryVersion(&xkb_lmaj, &xkb_lmin) && XkbQueryExtension(main_window->display, &xkb_opcode, &xkb_event, &xkb_error, &xkb_lmaj, &xkb_lmin))
+					XkbLockModifiers (main_window->display, XkbUseCoreKbd, LockMask, 0);
+			}
 			return;
 		}
 		case KLB_CLEAR:
@@ -792,6 +865,18 @@ static void program_perform_auto_action(struct _program *p, int action)
 		}
 		case KLB_DEL_SYM:
 		{
+			if (p->last_action == ACTION_AUTOCOMPLEMENTATION)
+			{
+				// Block events of keyboard (push to event queue)
+				set_event_mask(p->focus->owner_window, None);
+				
+				p->event->send_backspaces(p->event, 1);
+				p->last_action = ACTION_NONE;
+				
+				// Restore events mask
+				set_event_mask(p->focus->owner_window, INPUT_HANDLE_MASK | FOCUS_CHANGE_MASK | EVENT_KEY_MASK);
+			}
+			
 			string->del_symbol(string);
 			return;
 		}
@@ -799,15 +884,21 @@ static void program_perform_auto_action(struct _program *p, int action)
 		case KLB_SPACE:
 		case KLB_ADD_SYM:
 		{
-			if (action == KLB_ENTER && xconfig->dont_process_when_press_enter)
+			if (action == KLB_ENTER && xconfig->flush_buffer_when_press_enter)
+				p->buffer->save_and_clear(p->buffer, p->focus->owner_window);
+			
+			if (action == KLB_ENTER && xconfig->dont_process_when_press_enter && !xconfig->flush_buffer_when_press_enter)
 				action = KLB_ADD_SYM;
 			
+			if (p->changed_manual == MANUAL_FLAG_NEED_FLUSH)
+					p->changed_manual = MANUAL_FLAG_UNSET;
+			
 			char sym = main_window->keymap->get_cur_ascii_char(main_window->keymap, p->event->event);
 
 			if (action == KLB_ADD_SYM)
 			{
-				if (p->changed_manual == MANUAL_FLAG_NEED_FLUSH)
-					p->changed_manual = MANUAL_FLAG_UNSET;
+				// Correct small letter to capital letter after dot
+				p->check_capital_letter_after_dot(p);
 				
 				// Add symbol to internal bufer
 				int modifier_mask = groups[get_curr_keyboard_group()] | p->event->get_cur_modifiers(p->event);
@@ -816,6 +907,12 @@ static void program_perform_auto_action(struct _program *p, int action)
 				// Block events of keyboard (push to event queue)
 				set_event_mask(p->focus->owner_window, None);
 
+				// Correct space before punctuation
+			    p->check_space_before_punctuation(p);
+
+			    // Correct spaces with brackets
+			    p->check_space_with_bracket(p);
+				
 				p->check_brackets_with_symbols(p);
 				
 				if (!xconfig->check_lang_on_process)
@@ -862,6 +959,8 @@ static void program_perform_auto_action(struct _program *p, int action)
 			int modifier_mask = groups[get_curr_keyboard_group()] | p->event->get_cur_modifiers(p->event);
 			p->buffer->add_symbol(p->buffer, sym, p->event->event.xkey.keycode, modifier_mask);
 
+			//p->check_space_with_punctuation_mark(p);
+			
 			// Correct space before punctuation
 			p->check_space_before_punctuation(p);
 
@@ -876,8 +975,8 @@ static void program_perform_auto_action(struct _program *p, int action)
 			// Unblock keyboard
 			set_event_mask(p->focus->owner_window, INPUT_HANDLE_MASK | FOCUS_CHANGE_MASK | EVENT_KEY_MASK);
 
-			if (action == KLB_ENTER && xconfig->flush_buffer_when_press_enter)
-				p->buffer->save_and_clear(p->buffer, p->focus->owner_window);
+			//if (action == KLB_ENTER && xconfig->flush_buffer_when_press_enter)
+			//	p->buffer->save_and_clear(p->buffer, p->focus->owner_window);
 
 			p->last_action = ACTION_NONE;
 
@@ -1047,6 +1146,12 @@ static int program_perform_manual_action(struct _program *p, enum _hotkey_action
 			p->event->default_event.xkey.keycode = 0;
 			break;
 		}
+		case ACTION_ROTATE_LAYOUT_BACK:
+		{
+			set_prev_keyboard_group(xconfig->handle);
+			p->event->default_event.xkey.keycode = 0;
+			break;
+		}
 		case ACTION_AUTOCOMPLEMENTATION:
 		{
 			if (p->last_action == ACTION_AUTOCOMPLEMENTATION)
@@ -1081,6 +1186,54 @@ static int program_perform_manual_action(struct _program *p, enum _hotkey_action
 
 			break;
 		}
+		case ACTION_BLOCK_EVENTS:
+		{
+			p->buffer->save_and_clear(p->buffer, p->focus->owner_window);
+			p->event->default_event.xkey.keycode = 0;
+			xconfig->block_events = !xconfig->block_events;
+			if (xconfig->block_events)
+			{
+				grab_keyboard(p->focus->owner_window, TRUE);
+				show_notify(NOTIFY_BLOCK_EVENTS, NULL);
+			}
+			else
+			{
+				grab_spec_keys(p->focus->owner_window, TRUE);
+				show_notify(NOTIFY_UNBLOCK_EVENTS, NULL);
+			}
+			log_message (DEBUG, _("Now keyboard and mouse block status is %s"), xconfig->get_bool_name(xconfig->block_events));
+			
+			break;
+		}	
+		case ACTION_INSERT_DATE:
+		{
+			time_t curtime = time(NULL);
+			struct tm *loctime = localtime(&curtime);
+			if (loctime == NULL)
+				break;
+	
+			char *date = malloc(256 * sizeof(char));
+			strftime(date, 256, "%x", loctime);
+			
+			set_event_mask(p->focus->owner_window, None);
+			grab_spec_keys(p->focus->owner_window, FALSE);
+
+			// Insert Date 
+			log_message(DEBUG, _("Insert date '%s'."), date);
+
+			p->buffer->set_content(p->buffer, date);
+
+			p->change_word(p, CHANGE_INS_DATE);
+
+			set_event_mask(p->focus->owner_window, INPUT_HANDLE_MASK | FOCUS_CHANGE_MASK | EVENT_KEY_MASK);
+			grab_spec_keys(p->focus->owner_window, TRUE);
+
+			p->buffer->save_and_clear(p->buffer, p->focus->owner_window);
+
+			p->event->default_event.xkey.keycode = 0;
+			free (date);
+			break;
+		}
 		case ACTION_REPLACE_ABBREVIATION: // User needs to replace acronym
 		{
 			//MOVE this code to new function in new module
@@ -1157,12 +1310,46 @@ static int program_perform_manual_action(struct _program *p, enum _hotkey_action
 	}
 
 	// When CHANGE_STRING or CHANGE_WORD actions occured
-	if (xconfig->troubleshoot_switch)
-		p->changed_manual = MANUAL_FLAG_SET;
-
+	if ((xconfig->troubleshoot_switch) && (p->buffer->cur_pos > 0))
+	{
+		if (p->buffer->content[p->buffer->cur_pos-1] != ' ')
+			p->changed_manual = MANUAL_FLAG_SET;
+	}
 	return TRUE;
 }
 
+/* 
+static void get_suggest(struct _program *p, int lang)
+{
+	char *word = strdup(get_last_word(p->buffer->i18n_content[lang].content));
+	del_final_numeric_char(word);
+	
+	if (!enchant_dict_check(xconfig->handle->enchant_dicts[lang], word, strlen(word)))
+	{
+		free (word);
+		return;
+	}
+	
+	unsigned int count = 0;
+	char **suggs = enchant_dict_suggest (xconfig->handle->enchant_dicts[lang], word, strlen(word), &count); 
+	
+	log_message (ERROR, "Probably word \"%s\" is (count %d):", word, count);
+
+	for (unsigned int i = 0; i < count; i++)
+	{
+		log_message (ERROR, "    %d. %s", i+1, suggs[i]);
+	}
+	
+	free (word);
+
+	for (unsigned int i = 0; i < count; i++)
+	{
+		free (suggs[i]);
+	}
+	free (suggs);
+}
+*/
+
 static int program_check_lang_last_word(struct _program *p)
 {
 	if (xconfig->handle->languages[get_curr_keyboard_group()].excluded)
@@ -1178,18 +1365,28 @@ static int program_check_lang_last_word(struct _program *p)
 	if (!word)
 		return FALSE;
 
+	if ((p->buffer->content[p->buffer->cur_pos-1] == ' ') ||
+	    (p->buffer->content[p->buffer->cur_pos-1] == 0) || // Nothing word
+		(p->buffer->content[p->buffer->cur_pos-1] == 13) || // Enter
+	    (p->buffer->content[p->buffer->cur_pos-1] == 11))   // Tab
+			return FALSE;
+	
 	int cur_lang = get_curr_keyboard_group();
 	int new_lang = check_lang(xconfig->handle, p->buffer, cur_lang);
 
 	if (new_lang == NO_LANGUAGE)
 	{
 		log_message(DEBUG, _("No language found to change to"));
+		//get_suggest(p, cur_lang); 
 		return FALSE;
 	}
 
 	if (new_lang == cur_lang)
+	{
+		//get_suggest(p, cur_lang);
 		return FALSE;
-
+	}
+	
 	int change_action = CHANGE_WORD_TO_LAYOUT_0;
 	if (new_lang == 0)
 		change_action = CHANGE_WORD_TO_LAYOUT_0;
@@ -1202,6 +1399,10 @@ static int program_check_lang_last_word(struct _program *p)
 
 	p->change_word(p, change_action);
 	show_notify(NOTIFY_AUTOMATIC_CHANGE_WORD, NULL);
+
+	p->last_layout = new_lang;
+
+	//get_suggest(p, new_lang);
 	return TRUE;
 }
 
@@ -1247,6 +1448,9 @@ static int program_check_lang_last_syllable(struct _program *p)
 
 	p->change_word(p, change_action);
 	show_notify(NOTIFY_AUTOMATIC_CHANGE_WORD, NULL);
+
+	p->last_layout = new_lang;
+	
 	return TRUE;
 }
 
@@ -1281,10 +1485,13 @@ static void program_check_tcl_last_word(struct _program *p)
 
 	if (!isalpha(p->buffer->content[offset]))
 		return;
-
+	
 	if (p->buffer->cur_pos - offset <= 2)
 		return;
 
+	if (!isalpha(p->buffer->content[offset+1]))
+		return;
+	
 	if (isblank(p->buffer->content[offset+2]))
 		return;
 
@@ -1370,7 +1577,7 @@ static void program_check_space_with_bracket(struct _program *p)
 		return;
 	}
 	
-	if (((text[text_len - 1] == '(') && (text[text_len - 2] == ' ' || text[text_len - 2] == ':' || text[text_len - 2] == ';' || text[text_len - 2] == '-' || isdigit(text[text_len - 2]))) ||
+	if (((text[text_len - 1] == '(') && (text[text_len - 2] == ' ' || text[text_len - 2] == ':' || text[text_len - 2] == ';' || text[text_len - 2] == '-' || text[text_len - 2] == '\r' || text[text_len - 2] == '\n' || text[text_len - 2] == '\t' || isdigit(text[text_len - 2]))) ||
 	    ((text[text_len - 1] == ')' && text[text_len - 2] != ' ' )))
 	{
 		free(text);
@@ -1477,6 +1684,77 @@ static void program_check_brackets_with_symbols(struct _program *p)
 	free(text);
 }
 
+static void program_check_capital_letter_after_dot(struct _program *p)
+{
+	if (!xconfig->correct_capital_letter_after_dot)
+		return;
+
+	if (p->buffer->cur_pos < 3)
+		return;
+
+	if (p->event->event.xkey.state & ShiftMask)
+		return;
+
+	char *symbol = keycode_to_symbol(p->event->event.xkey.keycode, get_curr_keyboard_group(), p->event->event.xkey.state);
+	if (symbol == NULL)
+		return;
+
+	switch (symbol[0])
+	{
+		case '`':
+		case '1':
+		case '2':
+		case '3':
+		case '4':
+		case '5':
+		case '6':
+		case '7':
+		case '8':
+		case '9':
+		case '0':
+		case '-':
+		case '=':
+		case '[':
+		case ']':
+		case '\\':
+		case '/':
+		case ';':
+		case '\'':
+		case ',':
+		case '.':
+		{
+			free (symbol);
+			return;
+		}
+	}
+	
+	char *text = p->buffer->get_utf_string_on_kbd_group(p->buffer, get_curr_keyboard_group());
+	if (text == NULL)
+		return;
+
+	int offset = strlen(text) - 1;
+	if ((text[offset] != ' ') && (text[offset] != 13) && (text[offset] != 9))
+	{
+		free(text);
+		return;
+	}
+
+	for (offset = strlen(text) - 2; offset > 1; offset--)
+	{
+		if ((text[offset] != ' ') && (text[offset] != 13) && (text[offset] != 9))
+			break;
+	}
+	
+	if ((text[offset] == '.') || (text[offset] == '!') || (text[offset] == '?'))
+	{
+		log_message(DEBUG, _("Find small letter after dot, correction..."));
+		p->event->event.xkey.state = p->event->event.xkey.state | ShiftMask;
+		p->event->default_event.xkey.state = p->event->default_event.xkey.state | ShiftMask;
+	}
+	
+	free(text);
+}
+
 static void program_check_pattern(struct _program *p, int selection)
 {
 	if (!xconfig->autocomplementation)
@@ -1557,6 +1835,14 @@ static void program_send_string_silent(struct _program *p, int send_backspaces)
 
 	log_message(DEBUG, _("Processing string '%s'"), p->buffer->content);
 
+	// Work-arround
+	if ((xconfig->compatibility_with_completion) && (p->last_action != ACTION_AUTOCOMPLEMENTATION))
+	{
+		p->event->send_xkey(p->event, XKeysymToKeycode(main_window->display, XK_bar), 1);
+		p->event->send_backspaces(p->event, 1);
+		
+	} // end workarround
+	
 	p->event->send_backspaces(p->event, send_backspaces);		// Delete old string
 	p->event->send_string(p->event, p->buffer);		// Send new string
 }
@@ -1844,11 +2130,13 @@ static void program_change_word(struct _program *p, enum _change_action action)
 			break;
 		}
 		case CHANGE_ABBREVIATION:
+		case CHANGE_INS_DATE:
 		{
 			p->send_string_silent(p, 0);
 			break;
 		}
 	}
+
 	p->plugin->change_action(p->plugin, action);
 }
 
@@ -1871,11 +2159,16 @@ static void program_add_word_to_dict(struct _program *p, int new_lang)
 		return;
 	}
 
-	struct _list_char *curr_temp_dict = xconfig->handle->languages[curr_lang].temp_dict;
-	if (curr_temp_dict->exist(curr_temp_dict, curr_word, BY_PLAIN))
-		curr_temp_dict->rem(curr_temp_dict, curr_word);
-
-	struct _list_char *new_temp_dict = xconfig->handle->languages[new_lang].temp_dict;
+	struct _list_char *curr_temp_dictionary = xconfig->handle->languages[curr_lang].temp_dictionary;
+	if (curr_temp_dictionary->exist(curr_temp_dictionary, curr_word, BY_REGEXP))
+	{
+		char *word_to_dict = malloc((strlen(curr_word) + 7) * sizeof(char));
+		sprintf(word_to_dict, "%s%s%s", "(?i)^", curr_word, "$");
+		curr_temp_dictionary->rem(curr_temp_dictionary, word_to_dict);
+		free(word_to_dict);
+	}
+	
+	struct _list_char *new_temp_dictionary = xconfig->handle->languages[new_lang].temp_dictionary;
 
 	tmp = get_last_word(p->buffer->i18n_content[new_lang].content);
 
@@ -1889,28 +2182,37 @@ static void program_add_word_to_dict(struct _program *p, int new_lang)
 		return;
 	}
 
-	if (!new_temp_dict->exist(new_temp_dict, new_word, BY_PLAIN))
+	if (!new_temp_dictionary->exist(new_temp_dictionary, new_word, BY_REGEXP))
 	{
-		new_temp_dict->add(new_temp_dict, new_word);
+		char *word_to_dict = malloc((strlen(new_word) + 7) * sizeof(char));
+		sprintf(word_to_dict, "%s%s%s", "(?i)^", new_word, "$");
+		new_temp_dictionary->add(new_temp_dictionary, word_to_dict);
+		free(word_to_dict);
 		free(curr_word);
 		free(new_word);
 		return;
 	}
 
-	struct _list_char *curr_dict = xconfig->handle->languages[curr_lang].dict;
-	if (curr_dict->exist(curr_dict, curr_word, BY_PLAIN))
+	struct _list_char *curr_dictionary = xconfig->handle->languages[curr_lang].dictionary;
+	if (curr_dictionary->exist(curr_dictionary, curr_word, BY_REGEXP))
 	{
 		log_message(DEBUG, _("Remove word '%s' from %s dictionary"), curr_word, xconfig->handle->languages[curr_lang].name);
-		curr_dict->rem(curr_dict, curr_word);
+		char *word_to_dict = malloc((strlen(curr_word) + 7) * sizeof(char));
+		sprintf(word_to_dict, "%s%s%s", "(?i)^", curr_word, "$");
+		curr_dictionary->rem(curr_dictionary, word_to_dict);
 		xconfig->save_dict(xconfig, curr_lang);
+		free(word_to_dict);
 	}
 
-	struct _list_char *new_dict = xconfig->handle->languages[new_lang].dict;
-	if (!new_dict->exist(new_dict, new_word, BY_PLAIN))
+	struct _list_char *new_dictionary = xconfig->handle->languages[new_lang].dictionary;
+	if (!new_dictionary->exist(new_dictionary, new_word, BY_REGEXP))
 	{
 		log_message(DEBUG, _("Add word '%s' in %s dictionary"), new_word, xconfig->handle->languages[new_lang].name);
-		new_dict->add(new_dict, new_word);
+		char *word_to_dict = malloc((strlen(new_word) + 7) * sizeof(char));
+		sprintf(word_to_dict, "%s%s%s", "(?i)^", new_word, "$");
+		new_dictionary->add(new_dictionary, word_to_dict);
 		xconfig->save_dict(xconfig, new_lang);
+		free(word_to_dict);
 	}
 
 	p->add_word_to_pattern(p, new_lang);
@@ -1973,25 +2275,24 @@ static void program_add_word_to_pattern(struct _program *p, int new_lang)
 	}
 
 #ifdef WITH_ASPELL
-	AspellConfig *spell_config = new_aspell_config();
-	aspell_config_replace(spell_config, "lang", xconfig->handle->languages[new_lang].dir);
-	AspellCanHaveError *possible_err = new_aspell_speller(spell_config);
-
-	if (aspell_error_number(possible_err) != 0)
+	if (xconfig->handle->has_spell_checker[new_lang])
 	{
-		free(new_word);
-		delete_aspell_config(spell_config);
-		return;
+		if (!aspell_speller_check(xconfig->handle->spell_checkers[new_lang], new_word, strlen(new_word)))
+		{
+			free(new_word);
+			return;
+		}
 	}
-	
-	AspellSpeller *spell_checker = to_aspell_speller(possible_err);
-	int correct = aspell_speller_check(spell_checker, new_word, strlen(new_word));
-	delete_aspell_speller(spell_checker);
-	delete_aspell_config(spell_config);
-	if (!correct)
+#endif
+
+#ifdef WITH_ENCHANT
+	if (xconfig->handle->has_enchant_checker[new_lang])
 	{
-		free(new_word);
-		return;
+		if (enchant_dict_check(xconfig->handle->enchant_dicts[new_lang], new_word, strlen(new_word)))
+		{
+			free(new_word);
+			return;
+		}
 	}
 #endif
 	
@@ -2064,6 +2365,7 @@ struct _program* program_init(void)
 	p->check_space_before_punctuation	= program_check_space_before_punctuation;
 	p->check_space_with_bracket	= program_check_space_with_bracket;
 	p->check_brackets_with_symbols = program_check_brackets_with_symbols;
+	p->check_capital_letter_after_dot = program_check_capital_letter_after_dot;
 	p->check_pattern	= program_check_pattern;
 	p->change_word			= program_change_word;
 	p->add_word_to_dict		= program_add_word_to_dict;
